"""
Utilities for tests.
"""

import os
from contextlib import contextmanager


def appender(target, *args):
    """
    A function that appends given arguments and its own to the specified
    list.
    """

    def append(*append_args):
        """Append the earlier specified and given arguments to the target."""
        target.append(args + append_args)

    return append


def before_after(before, after):
    """
    A context manager calling given functions before and after the context,
    respectively, with the same arguments as it is given itself.
    """

    @contextmanager
    def around(*args, **kwargs):
        """Call given functions before and after the context."""
        before(*args, **kwargs)
        yield
        after(*args, **kwargs)

    return around


@contextmanager
def set_environ(**kwargs):
    """
    A context manager setting environment variable to given values before the
    context and restoring them afterwards.
    """

    old_values = {}
    for key, value in kwargs.items():
        old_values[key] = os.environ.get(key, None)
        os.environ[key] = value

    try:
        yield

    finally:
        for key, value in old_values.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value



"""
Test running Aloe in a separate process, as the real users would.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import re
import subprocess
import sys
import unittest

from aloe.testing import in_directory

from tests.utils import set_environ


TEST_PATH = os.path.dirname(__file__)

ROOT_PATH = os.path.dirname(os.path.dirname(TEST_PATH))


@in_directory('tests/simple_app')
class SimpleIntegrationTest(unittest.TestCase):
    """
    Test that basic feature running works as expected.
    """

    def test_success(self):
        """
        Test running a simple feature.
        """

        exitcode, _ = self.run_feature('features/calculator.feature')
        self.assertEqual(exitcode, 0, "Feature run successfully.")

    def test_success_zh(self):
        """
        Test running a simple feature with non-ASCII characters, with verbose
        output.
        """

        exitcode, out = self.run_feature('features/calculator_zh.feature',
                                         '--verbosity=3',
                                         terminal=True)

        self.assertEqual(exitcode, 0, "Feature run successfully.")

        # Remove timing information from the output as unstable
        out = re.sub(b'in [0-9.]+s', b'in XXXXs', out)

        with open(os.path.join(TEST_PATH, 'calculator.txt'), 'rb') as expected:
            expected_out = expected.read()

        if out != expected_out:
            actual_path = os.path.join(TEST_PATH, 'calculator-actual.txt')
            with open(actual_path, 'wb') as actual:
                actual.write(out)

        self.assertEqual(out, expected_out, "Output matches expected.")

    def test_failure(self):
        """
        Test running a failing feature.
        """

        exitcode, _ = self.run_feature('features/wrong_expectations.feature')
        self.assertNotEqual(exitcode, 0, "Feature run failed.")

    def run_feature(self, *args, **kwargs):
        """
        Run a feature and return the (exitcode, output) tuple.
        :param args: Arguments to pass to the Aloe subprocess
        :param terminal: Whether to run in a terminal
        :returns: (exit code, output)
        """

        # Python 2 doesn't support named kwargs after star-args
        terminal = kwargs.pop('terminal', False)
        if kwargs:
            raise TypeError("Invalid arguments.")

        args = [sys.executable, '-c', 'import aloe; aloe.main()'] + list(args)

        # Ensure Aloe itself is on the path
        with set_environ(PYTHONPATH=ROOT_PATH):

            if terminal:
                try:
                    import pty
                except ImportError:
                    raise unittest.SkipTest("PTY support unavailable.")

                chunks = [b'']

                def read(file_desc):
                    """Store the subprocess output."""
                    data = os.read(file_desc, 1024)
                    chunks.append(data)
                    return data

                with set_environ(
                        COLORTERM='',
                        TERM='xterm-256color',
                        TERM_PROGRAM='',
                ):

                    status = pty.spawn(args, read)  # pylint:disable=assignment-from-no-return

                    # On Python 2, pty.spawn doesn't return the exit code
                    if status is None:
                        (_, status) = os.wait()  # pylint:disable=no-member

                    return status, b''.join(chunks)

            try:
                output = subprocess.check_output(
                    args,
                    stderr=subprocess.STDOUT,
                )
                return 0, output
            except subprocess.CalledProcessError as ex:
                return ex.returncode, ex.output



# -*- coding: utf-8 -*-
"""
Steps for testing the basic Gherkin test functionality.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import super
# pylint:enable=redefined-builtin

from contextlib import contextmanager

from nose.tools import assert_equal

from aloe import (
    after,
    around,
    before,
    step,
    world,
)
from aloe.testclass import TestCase


class CallbackTestCase(TestCase):  # pylint:disable=abstract-method
    """Base for Gherkin test cases recording the use of setUp()/tearDown()."""

    def setUp(self):
        """Emit a 'setUp' event."""

        super().setUp()
        record_event('testclass', '[')

    def tearDown(self):
        """Emit a 'tearDown' event."""

        record_event('testclass', ']')
        super().tearDown()


def record_event(kind, value):
    """
    Record an event of a particular kind. Used for testing the order of
    callback execution.
    """

    kind = kind.replace('"', '')

    if not hasattr(world, kind):
        setattr(world, kind, [])
    events = getattr(world, kind)

    events.append(value)


@before.each_step
def before_step(self):
    """Record the "before step" event."""
    record_event('step', '{')
    record_event('step_names', ('before', self.sentence))
    record_event('step_testclasses', self.testclass)
    record_event('step_tests', self.test)
    record_event('types', 'before step')


@around.each_step
@contextmanager
def around_step(self):
    """Record the "around step" event."""
    record_event('step', '[')
    record_event('step_names', ('around', self.sentence))
    record_event('types', 'around_before step')
    yield
    record_event('types', 'around_after step')
    record_event('step', ']')


@after.each_step
def after_step(self):
    """Record the "after step" event."""
    record_event('types', 'after step')
    record_event('step', '}')
    record_event('step_names', ('after', self.sentence))


def record_example_event(when, scenario, outline, steps):
    """Record a scenario- or scenario outline-level event."""
    if outline:
        result = "Outline: " + scenario.name
        result += ' (' + \
            ', '.join('='.join((k, v)) for k, v in outline.items()) + \
            ')'
    else:
        result = "Scenario: " + scenario.name

    result += ", steps={}".format(len(steps))
    record_event('example_names', (when, result))


@before.each_example
def before_example(scenario, outline, steps):
    """Record the "before example" event."""
    record_event('example', '{')
    record_example_event('before', scenario, outline, steps)
    record_event('types', 'before example')


@around.each_example
@contextmanager
def around_example(scenario, outline, steps):
    """Record the "around example" event."""
    record_event('example', '[')
    record_example_event('around', scenario, outline, steps)
    record_event('types', 'around_before example')
    yield
    record_event('types', 'around_after example')
    record_event('example', ']')


@after.each_example
def after_example(scenario, outline, steps):
    """Record the "after example" event."""
    record_event('types', 'after example')
    record_event('example', '}')
    record_example_event('after', scenario, outline, steps)


@before.each_feature
def before_feature(feature):
    """Record the "before feature" event."""
    record_event('feature', '{')
    record_event('feature_names', ('before', feature.name))
    record_event('feature_testclasses', feature.testclass)
    record_event('types', 'before feature')


@around.each_feature
@contextmanager
def around_feature(feature):
    """Record the "around feature" event."""
    record_event('feature', '[')
    record_event('feature_names', ('around', feature.name))
    record_event('types', 'around_before feature')
    yield
    record_event('types', 'around_after feature')
    record_event('feature', ']')


@after.each_feature
def after_feature(feature):
    """Record the "after feature" event."""
    record_event('types', 'after feature')
    record_event('feature', '}')
    record_event('feature_names', ('after', feature.name))


@before.all
def before_all():
    """Record the "before all" event."""
    record_event('"all"', '{')
    record_event('types', 'before all')


@around.all
@contextmanager
def around_all():
    """Record the "around all" event."""
    record_event('"all"', '[')
    record_event('types', 'around_before all')
    yield
    record_event('types', 'around_after all')
    record_event('"all"', ']')


@after.all
def after_all():
    """Record the "after all" event."""
    record_event('types', 'after all')
    record_event('"all"', '}')


@step(r'I emit an? ([^ ]+) event of "([^"]+)"')
@step(r'Я записываю событие ([^ ]+) "([^"]+)"')
def emit_event(self, kind, event):
    """Record a user-defined event."""
    record_event(kind, event)


@step(r'I emit an? ([^ ]+) event for each letter in "([^"]+)"')
def emit_event_letters(self, kind, letters):
    """
    Record several user-defined events in succession.

    Used to test behave_as.
    """

    for letter in letters:
        self.when('I emit a {kind} event of "{letter}"'.format(
            kind=kind,
            letter=letter,
        ))


@step(r'Я записываю событие ([^ ]+) для каждой буквы в "([^"]+)"')
def emit_event_letters_ru(self, kind, letters):
    """
    Record several user-defined events in succession, using the Russian step
    definition.

    Used to test behave_as.
    """

    for letter in letters:
        self.when('Я записываю событие {kind} "{letter}"'.format(
            kind=kind,
            letter=letter,
        ))


@step(r'The ([^ ]+) event sequence should be "([^"]+)"')
@step(r'Последовательность событий ([^ ]+) должна быть "([^"]+)"')
def check_events(self, kind, events):
    """Check the recorded events of a particular type."""
    kind = kind.replace('"', '')
    assert_equal(''.join(getattr(world, kind)), events)


@after.each_step
def count_failures_successes(self):
    """Count the number of passed and failed steps."""
    if self.failed:
        world.failed_steps += 1
    if self.passed:
        world.passing_steps += 1


@before.each_example
def reset_failures_successes_count(scenario, outline, steps):
    """Reset the passed/failed step count for each scenario."""
    world.passing_steps = 0
    world.failed_steps = 0


@step(r'I have a passing step')
def good_step(self):
    """A step that always passes."""
    pass


@step(r'I have a failing step')
def bad_step(self):
    """A step that always fails."""
    assert False, "This step is meant to fail."



"""
Steps for testing running multiple apps together.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from aloe import (
    before,
    step,
    world,
)


@before.all
def start_two():
    """Mark the app two as started."""
    world.app_two_started = True


@step(r'I record the started callbacks in app two')
def check_callbacks_two(self):
    """Check which apps were started."""
    world.started_callbacks_two = {
        'one': getattr(world, 'app_one_started', None),
        'two': getattr(world, 'app_two_started', None),
    }



"""
Cucumber-esque outputter for Nose.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import

import os
import sys
from contextlib import contextmanager
from functools import wraps

import colorama
from colorama import Cursor
import colors

from aloe.registry import (
    CallbackDecorator,
    CALLBACK_REGISTRY,
    PriorityClass,
)
from aloe.strings import ljust, represent_table
from aloe.tools import hook_not_reentrant
from aloe.utils import memoizedproperty, PY3
from nose.result import TextTestResult

# A decorator to add callbacks which wrap the steps looser than all the other
# callbacks.
# pylint:disable=invalid-name
outer_around = CallbackDecorator(CALLBACK_REGISTRY, 'around',
                                 priority_class=PriorityClass.DISPLAY)
# pylint:enable=invalid-name


# Initialize Colorama to support colored output on Windows.
colorama.init()

# Global reference to the Terminal:
# This exists because the hooks have to be registered before the test is
# started, which is when the stream is passed in.
TERMINAL = [None]


class Terminal(object):
    """
    Output coloring and movement.
    """

    # Coloring of various output parts
    DEFAULT_THEME = {
        'preview': 'grey',
        'pending': 'yellow',
        'failed': 'red',
        'passed': 'green',
        'skipped': 'cyan',
        'comment': 'grey',
        'tag': 'cyan',
    }

    @memoizedproperty
    def theme(self):
        """The color theme, taking CUCUMBER_COLORS into account."""

        theme = self.DEFAULT_THEME.copy()
        for theme_element in os.environ.get('CUCUMBER_COLORS', '').split(':'):
            try:
                element, color = theme_element.split('=')
            except ValueError:
                # Ignore invalid values
                continue

            theme[element] = color

        return theme

    is_a_tty = sys.stdout.isatty()

    move_up = Cursor.UP()

    def __init__(self, kind=None, stream=None, force_styling=False):
        if stream is None:
            stream = sys.__stdout__
        self.stream = stream
        self.does_styling = self.is_a_tty or force_styling

    def __nonzero__(self):
        return True

    def __getattr__(self, attr):
        """Create and return color methods for coloring output."""

        color = self.theme[attr]  # pylint:disable=unsubscriptable-object

        return self.colored(color)

    def colored(self, color):
        """A function to output a string in the given color."""

        if not self.does_styling:
            return lambda s: s

        # Grey is not one of the basic 16 colors
        if color == 'grey':
            color = 243

        return lambda s: colors.color(s, fg=color)

    def write(self, arg='', return_=False):
        """
        Write to the stream.

        :param return_: if True return the cursor back to where it was
            before the write.
        """

        # On Python 2, the default output encoding is ASCII
        self.stream.write(arg if PY3 else arg.encode('utf-8'))

        if return_:
            self.stream.write(self.move_up * arg.count('\n'))

    def writeln(self, arg='', return_=False):
        """Convenience function to write a line to the stream."""
        self.write(arg + '\n', return_=return_)

    @classmethod
    def required(cls, func):
        """Decorate this hook to only execute if term is available"""

        @wraps(func)
        def inner(*args, **kwargs):
            """Wrap func to check for term."""

            def empty_generator():
                """
                Hide the generator in a separate function
                because Python 2 can't support "returning from generators"
                """
                yield

            if not TERMINAL[0]:
                return empty_generator()
            else:
                return func(TERMINAL[0], *args, **kwargs)

        return inner


@outer_around.each_feature
@contextmanager
@Terminal.required
def feature_wrapper(term, feature):
    """Display feature execution."""

    try:
        if feature.tags:
            term.writeln(term.tag(feature.represent_tags()))

        lines = feature.represented().splitlines()
        term.writeln(lines[0])
        term.writeln()
        term.writeln('\n'.join(lines[1:]))
        term.writeln()

        yield
    finally:
        pass


@outer_around.each_example
@contextmanager
@Terminal.required
def example_wrapper(term, scenario, outline, steps):
    """Display scenario execution."""

    try:
        if scenario.tags:
            term.writeln(term.tag(scenario.represent_tags()))

        represented = scenario.represented()
        represented = ljust(represented, scenario.feature.max_length + 2)

        term.write(represented)
        term.writeln(term.comment('# ' + scenario.location))

        if outline:
            term.writeln(represent_table([outline.keys(),
                                          outline.values()],
                                         indent=6))
            term.writeln()

        # write a preview of the steps
        if term.is_a_tty:
            steps_ = []

            if scenario.feature.background:
                steps_ += scenario.feature.background.steps

            steps_ += steps

            term.writeln(
                term.preview('\n'.join(
                    step.represented()
                    for step in steps_
                ) + '\n'),
                return_=True
            )

        yield
    finally:
        term.writeln()


@outer_around.each_step
@contextmanager
@Terminal.required
@hook_not_reentrant  # don't display inner steps called by top-level steps
def step_wrapper(term, step):
    """Display step execution."""

    try:
        if term.is_a_tty:
            term.writeln(
                step.represented(color=term.pending),
                return_=True)

        yield
    finally:
        if step.passed:
            color = term.passed
        elif step.failed:
            color = term.failed
        else:
            color = term.skipped

        term.writeln(
            step.represented(color=color)
        )


class AloeTestResult(TextTestResult):
    """
    Cucumber test progress display (verbosity level 3).
    """
    # pylint:disable=too-many-arguments
    def __init__(self, stream, descriptions, verbosity,
                 config=None, errorClasses=None):
        super().__init__(stream, descriptions, verbosity,
                         config=config, errorClasses=errorClasses)
        self.showAll = verbosity == 2
        self.showSteps = verbosity >= 3  # pylint:disable=invalid-name

        if self.showSteps:
            # Register the global terminal so that it can be accessed by
            # the hooks.
            TERMINAL[0] = Terminal(
                stream=stream,
                force_styling=config.force_color)
        else:
            # unset the global -- primarily for the tests where we're going
            # to create more TestResults with different streams.
            TERMINAL[0] = None



# -*- coding: utf-8 -*-
"""
Test utilities from aloe.strings.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from nose.tools import assert_equal

from aloe import strings


def test_represent_table():
    """
    Test representing a table
    """

    table = [
        ['name', 'age'],
        [u'Gabriel Falcão', 22],
        ['Miguel', 19],
    ]

    assert_equal(
        strings.represent_table(table),
        u"| name           | age |\n"
        u"| Gabriel Falcão | 22  |\n"
        u"| Miguel         | 19  |"
    )


def test_represent_table_escapes_pipe():
    """
    Test representing a table with escaping
    """

    table = [
        ['name', 'age'],
        [u'Gabriel | Falcão', 22],
        ['Miguel | Arcanjo', 19],
    ]

    assert_equal(
        strings.represent_table(table),
        '\n'.join((
            r"| name              | age |",
            r"| Gabriel \| Falcão | 22  |",
            r"| Miguel \| Arcanjo | 19  |",
        ))
    )


def test_represent_table_allows_empty():
    """
    Test representing a table with an empty cell
    """

    table = [
        ['name', 'age'],
        [u'Gabriel | Falcão', 22],
        ['Miguel | Arcanjo', ''],
    ]

    assert_equal(
        strings.represent_table(table),
        '\n'.join((
            r"| name              | age |",
            r"| Gabriel \| Falcão | 22  |",
            r"| Miguel \| Arcanjo |     |",
        ))
    )


def test_column_width():
    """strings.column_width"""

    assert_equal(
        strings.get_terminal_width(u"あいうえお"),
        10
    )


def test_column_width_w_number_and_char():
    """strings.column_width_w_number_and_char"""

    assert_equal(
        strings.get_terminal_width(u"%s%c" % (u"4209", 0x4209)),
        6
    )



"""
Base test class for tests generated from scenarios.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import zip
from builtins import str
from builtins import range
from builtins import super
# pylint:enable=redefined-builtin

import ast
import unittest
from contextlib import contextmanager

from nose.plugins.attrib import attr

from aloe.codegen import make_function
from aloe.fs import path_to_module_name
from aloe.parser import (
    Background,
    Feature,
    Scenario,
    Step,
)
from aloe.registry import (
    CallbackDecorator,
    CALLBACK_REGISTRY,
    PriorityClass,
    STEP_REGISTRY,
)
from aloe.utils import identifier


class TestStep(Step):
    """
    A step with additional functions for the callbacks.
    """

    @property
    def testclass(self):
        """
        The test class containing this step (in a scenario or a background).
        """
        return self.feature.testclass

    test = None
    """
    The test currently running the step, or None if not currently in a test
    (e.g. in a `before_feature` callback).
    """

    def __init__(self, *args, **kwargs):
        """Initialize the step status."""
        self.failed = None
        self.passed = None
        super().__init__(*args, **kwargs)

    def behave_as(self, string):
        """Run the specified step in the current context."""
        self.test.behave_as(self, string)

    def given(self, string):
        """Run the specified 'Given' step in the current context."""
        self.behave_as(self.step_keyword('given') + string)

    def when(self, string):
        """Run the specified 'When' step in the current context."""
        self.behave_as(self.step_keyword('when') + string)

    def then(self, string):
        """Run the specified 'Then' step in the current context."""
        self.behave_as(self.step_keyword('then') + string)


class TestBackground(Background):
    """A background creating steps for testing."""
    step_class = TestStep


class TestScenario(Scenario):
    """A background creating steps for testing."""
    step_class = TestStep


class TestFeature(Feature):
    """A feature creating steps for testing."""
    background_class = TestBackground
    scenario_class = TestScenario


class TestCase(unittest.TestCase):
    """
    The base test class for tests compiled from Gherkin features.
    """

    feature = None  # Will be supplied when constructing derived classes

    @classmethod
    def before_feature(cls, feature):
        """Call feature-level before callbacks."""
        raise NotImplementedError(
            "This should be supplied when constructing derived classes.")

    @classmethod
    def after_feature(cls, feature):
        """Call feature-level after callbacks."""
        raise NotImplementedError(
            "This should be supplied when constructing derived classes.")

    # Methods for the use of the tested code

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.before_feature(cls.feature)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.after_feature(cls.feature)

    def behave_as(self, context_step, string):
        """
        Run the steps described by the given string in the context of the
        step.
        """

        steps = context_step.parse_steps_from_string(string)

        # Copy necessary attributes onto new steps
        for step in steps:
            step.test = self

            try:
                step.scenario = context_step.scenario
            except AttributeError:
                step.background = context_step.background

            definition = self.prepare_step(step)

            definition['func'](definition['step'],
                               *definition['args'],
                               **definition['kwargs'])

    def shortDescription(self):
        return str(self)

    def __str__(self):
        test_method = getattr(self, self._testMethodName)
        return "%s (%s: %s)" % (
            test_method.scenario.name,
            path_to_module_name(self.feature.filename),
            self.feature.name,
        )

    # Methods for generating test classes

    @classmethod
    def from_file(cls, file_):
        """
        Construct a test class from a feature file.
        """

        feature = TestFeature.from_file(file_)

        background = cls.make_background(feature.background)
        scenarios = [
            example
            for i, scenario in enumerate(feature.scenarios)
            for example in cls.make_examples(scenario, i + 1)
        ]

        before_feature, after_feature = \
            CALLBACK_REGISTRY.before_after('feature')

        members = {
            'feature': feature,
            'background': background,
            'before_feature': staticmethod(before_feature),
            'after_feature': staticmethod(after_feature),
        }

        members.update({
            scenario.__name__: scenario
            for scenario in scenarios
        })

        class_name = identifier(feature.name)

        testclass = type(class_name, (cls,), members)
        testclass.feature.testclass = testclass
        return testclass

    @classmethod
    def make_background(cls, background):
        """
        Construct a method running the background steps.
        """

        if background is None:
            result = make_function('def background(self): pass')
        else:
            result = cls.make_steps(background,
                                    background.steps,
                                    is_background=True)

        return result

    @classmethod
    def make_examples(cls, scenario, index):
        """
        Construct methods for running all the examples of a scenario.

        index is the 1-based number of the scenario in the feature.
        """

        if scenario.outlines:
            for i, (outline, steps) in enumerate(scenario.evaluated, 1):
                # Create a function calling the real scenario example to show
                # the right location in the outline
                source = """
def run_example(self):
    outline(self)
                """
                source = ast.parse(source)

                # Set location of the call
                for node in ast.walk(source.body[0].body[0]):
                    node.lineno = outline.line

                context = {
                    'outline': cls.make_steps(scenario,
                                              steps,
                                              is_background=False,
                                              outline=outline)
                }

                yield cls.make_example(
                    make_function(
                        source=source,
                        context=context,
                        source_file=scenario.feature.filename,
                        name='{}: Example {}'.format(scenario.name, i),
                    ),
                    scenario,
                    index,
                )
        else:
            yield cls.make_example(
                cls.make_steps(
                    scenario,
                    scenario.steps,
                    is_background=False,
                ),
                scenario,
                index,
            )

    @classmethod
    def make_example(cls, method, scenario, index):
        """
        Set the method attributes to associate it with given scenario and index.
        """

        method.is_example = True
        method.scenario = scenario
        method.scenario_index = index

        for tag in scenario.tags:
            method = attr(tag)(method)

        return method

    @classmethod
    def prepare_step(cls, step):
        """
        Find a definition for the step.

        Returns a dictionary of: step, func, args, kwargs, where:
        - step is the original step
        - func is the function to run (wrapped in callbacks)
        - args and kwargs are the arguments to pass to the function
        """

        func, args, kwargs = STEP_REGISTRY.match_step(step)
        func = CALLBACK_REGISTRY.wrap('step', func, step)

        return {
            'step': step,
            'func': func,
            'args': args,
            'kwargs': kwargs,
        }

    @classmethod
    def make_steps(cls, step_container, steps,
                   is_background, outline=None):
        """
        Construct either a scenario or a background calling the specified
        steps.

        The method will have debugging information corresponding to the lines
        in the feature file.
        """

        assert steps

        step_definitions = [
            cls.prepare_step(step)
            for step in steps
        ]

        source = 'def run_steps(self):\n'
        if not is_background:
            source += '    self.background()\n'
        source += '\n'.join(
            # This has to be a single statement, in order to set its source
            # location as a whole below
            """
    try:
        step{i}.test = self
        func{i}(step{i}, *args{i}, **kwargs{i})
    finally:
        step{i}.test = None
            """.format(i=i)
            for i in range(len(step_definitions))
        )
        source = ast.parse(source)

        # Set locations of the steps
        step_source = source.body[0].body
        if not is_background:
            # There is no source for the background() call
            step_source = step_source[1:]
        for step, step_call in zip(steps, step_source):
            for node in ast.walk(step_call):
                node.lineno = step.line

        # Supply all the step functions and arguments
        context = {
            k + str(i): v
            for i, definition in enumerate(step_definitions)
            for k, v in definition.items()
        }

        if is_background:
            func_name = 'background'
        else:
            func_name = step_container.name

        run_steps = make_function(
            source=source,
            context=context,
            source_file=step_container.filename,
            name=func_name,
        )

        if not is_background:
            run_steps = CALLBACK_REGISTRY.wrap('example', run_steps,
                                               step_container, outline, steps)

        return run_steps

    @classmethod
    def scenarios(cls):
        """
        Enumerate all the scenario method names in the test class, in the
        order they were declared in the feature, together with their indices.
        """

        scenarios = {
            name: method
            for name, method in cls.__dict__.items()
            if getattr(method, 'is_example', False)
        }

        with_indices = [
            (method.scenario_index, name)
            for name, method in scenarios.items()
        ]

        return sorted(with_indices)


# A decorator to add callbacks which wrap the steps tighter than all the user
# callbacks.
# pylint:disable=invalid-name
inner_around = CallbackDecorator(CALLBACK_REGISTRY, 'around',
                                 priority_class=PriorityClass.SYSTEM_INNER)
# pylint:enable=invalid-name


@inner_around.each_step
@contextmanager
def set_passed_failed(step):
    """
    Set the 'failed' property of the step.
    """

    try:
        yield
        step.passed = True
        step.failed = False
    except:
        step.passed = False
        step.failed = True
        raise
